! -------------------------------------------------------------------
!{*************Uniform random number generator**********************}
!
!
        double precision function uniform()
        IMPLICIT NONE
!
!       Generate uniformly distributed random numbers using the 32-bit
!       generator from figure 3 of:
!       L'Ecuyer, P. Efficient and portable combined random number
!       generators, C.A.C.M., vol. 31, 742-749 & 774-?, June 1988.
!
!       The cycle length is claimed to be 2.30584E+18
!
!       Seeds can be set by calling the routine set_uniform
!
!       It is assumed that the Fortran compiler supports long variable
!       names, and integer*4.
!
        integer*4 z, k, is1, is2
        common /unif_seeds/ is1, is2
        save /unif_seeds/
!
        k = is1 / 53668
        is1 = 40014 * (is1 - k * 53668) - k * 12211
        if (is1 .lt. 0) then
           is1 = is1 + 2147483563
        endif
!
        k = is2 / 52774
        is2 = 40692 * (is2 - k * 52774) - k * 3791
        if (is2 .lt. 0) then
           is2 = is2 + 2147483399
        endif
!
        z = is1 - is2
        if (z .lt. 1) then
           z = z + 2147483562
        endif
!
        uniform = z / 2147483563.d0
        return
        end

!--------------------------------------------------------------------
!{**************** normal random number generator ******************}
 
        DOUBLE PRECISION FUNCTION ran_nor(m,v)
        IMPLICIT NONE
        DOUBLE PRECISION m, v, uniform, pi,z,huge
        INTEGER*4 is1, is2
        common /unif_seeds/ is1, is2
        save /unif_seeds/
 
        pi = 3.1415926536d0
       huge = 1e35
 1     z=dsqrt(-2.d0*dlog(uniform()))*dcos(2.d0*pi*uniform())
       if (dabs(z).ge.huge) goto 1
       ran_nor = m + sqrt(v)*z
       return
       end

!--------------------------------------------------------------------
!{***************** gamma random number generator ******************}

          double precision function ran_gam(aa, bb)
          IMPLICIT NONE
          double precision aa, bb, d, c, w, u, v, x, z, yy, e,uniform
          logical acc
          integer*4 is1, is2
          common /unif_seeds/ is1, is2
          save /unif_seeds/
!         integer*4 l
          

          if (aa.lt.1.d0) then
!               { ********************* Johnk's algorithm***}
         x = 2.d0
         yy=0.d0
         do while((x+yy).gt.1.d0)
           u = uniform()
           v = uniform()
           x = exp(log(u)/aa)
           yy = exp(log(v)/(1.d0-aa))
           enddo
         e = -log(uniform())
         z = e*x/(x + yy)
         ran_gam = z * bb

        else if (aa.eq.1.d0) then
        ran_gam = -bb * log(uniform())
        else
!               { ************************* Best's algorithm***}
        d = aa-1.d0
        c = 3.d0*aa - (.75d0)
        acc = .false.
        do while (.not.acc)
          u = uniform()
          v = uniform()
          w = u*(1.d0-u)
          yy = sqrt(c/w)*(u-0.5d0)
          x = d+yy
          if (x.gt.0.d0) then
            z = 64.d0*w**3*v**2
            acc = (z.le.(1.d0-2.d0*yy**2/x))
            if (.not.acc) acc = (log(z).le.(2.d0 * (d*log(x/d)-yy) ))
            endif
          enddo
        ran_gam = x * bb
       endif
       return
       end

!--------------------------------------------------------------------
!{************ inverse gamma random number generator ***************}
      double precision function ran_invgam(del, lam)
      IMPLICIT NONE
      REAL *8 :: del, lam, ran_gam

      INTEGER*4 is1, is2
      common /unif_seeds/ is1, is2
      save /unif_seeds/
	  
      ran_invgam = 1.0d0/ran_gam(del, 1.0d0/lam)

      end
	  
!--------------------------------------------------------------------
!{** Set seeds for the uniform random number generator. ************}

      SUBROUTINE set_uniform(iseed1, iseed2)

      IMPLICIT NONE
      INTEGER :: is1, is2, iseed1, iseed2
      COMMON /unif_seeds/ is1, is2
      SAVE /unif_seeds/

      is1 = iseed1
      is2 = iseed2
      RETURN
      END SUBROUTINE set_uniform
	  
! -------------------------------------------------------------------
!{**************************** SORTING ******************************}
! From Leonard J. Moss of SLAC:
!http://ftp.tuwien.ac.at/languages/fortran/freemarket/quick_sort2.f
! Here's a hybrid QuickSort I wrote a number of years ago.  It's
! based on suggestions in Knuth, Volume 3, and performs much better
! than a pure QuickSort on short or partially ordered input arrays.  

      SUBROUTINE SORTRX(N,DATA,INDEX)
!===================================================================
!
!     SORTRX -- SORT, Real input, indeX output
!
!
!     Input:  N     INTEGER
!             DATA  REAL
!
!     Output: INDEX INTEGER (DIMENSION N)
!
! This routine performs an in-memory sort of the first N elements of
! array DATA, returning into array INDEX the indices of elements of
! DATA arranged in ascending order.  Thus,
!
!    DATA(INDEX(1)) will be the smallest number in array DATA;
!    DATA(INDEX(N)) will be the largest number in DATA.
!
! The original data is not physically rearranged.  The original order
! of equal input values is not necessarily preserved.
!
!===================================================================
!
! SORTRX uses a hybrid QuickSort algorithm, based on several
! suggestions in Knuth, Volume 3, Section 5.2.2.  In particular, the
! "pivot key" [my term] for dividing each subsequence is chosen to be
! the median of the first, last, and middle values of the subsequence;
! and the QuickSort is cut off when a subsequence has 9 or fewer
! elements, and a straight insertion sort of the entire array is done
! at the end.  The result is comparable to a pure insertion sort for
! very short arrays, and very fast for very large arrays (of order 12
! micro-sec/element on the 3081K for arrays of 10K elements).  It is
! also not subject to the poor performance of the pure QuickSort on
! partially ordered data.
!
! Created:  15 Jul 1986  Len Moss
!
!===================================================================
 
      INTEGER   N,INDEX(N)
      REAL *8      DATA(N)
 
      INTEGER   LSTK(31),RSTK(31),ISTK
      INTEGER   L,R,I,J,P,INDEXP,INDEXT
      REAL *8     DATAP
 
!     QuickSort Cutoff
!
!     Quit QuickSort-ing when a subsequence contains M or fewer
!     elements and finish off at end with straight insertion sort.
!     According to Knuth, V.3, the optimum value of M is around 9.
 
      INTEGER   M
      PARAMETER (M=9)
 
!===================================================================
!
!     Make initial guess for INDEX
 
      DO 50 I=1,N
         INDEX(I)=I
   50    CONTINUE
 
!     If array is short, skip QuickSort and go directly to
!     the straight insertion sort.
 
      IF (N.LE.M) GOTO 900
 
!===================================================================
!
!     QuickSort
!
!     The "Qn:"s correspond roughly to steps in Algorithm Q,
!     Knuth, V.3, PP.116-117, modified to select the median
!     of the first, last, and middle elements as the "pivot
!     key" (in Knuth's notation, "K").  Also modified to leave
!     data in place and produce an INDEX array.  To simplify
!     comments, let DATA[I]=DATA(INDEX(I)).
 
! Q1: Initialize
      ISTK=0
      L=1
      R=N
 
  200 CONTINUE
 
! Q2: Sort the subsequence DATA[L]..DATA[R].
!
!     At this point, DATA[l] <= DATA[m] <= DATA[r] for all l < L,
!     r > R, and L <= m <= R.  (First time through, there is no
!     DATA for l < L or r > R.)
 
      I=L
      J=R
 
! Q2.5: Select pivot key
!
!     Let the pivot, P, be the midpoint of this subsequence,
!     P=(L+R)/2; then rearrange INDEX(L), INDEX(P), and INDEX(R)
!     so the corresponding DATA values are in increasing order.
!     The pivot key, DATAP, is then DATA[P].
 
      P=(L+R)/2
      INDEXP=INDEX(P)
      DATAP=DATA(INDEXP)
 
      IF (DATA(INDEX(L)) .GT. DATAP) THEN
         INDEX(P)=INDEX(L)
         INDEX(L)=INDEXP
         INDEXP=INDEX(P)
         DATAP=DATA(INDEXP)
      ENDIF
 
      IF (DATAP .GT. DATA(INDEX(R))) THEN
         IF (DATA(INDEX(L)) .GT. DATA(INDEX(R))) THEN
            INDEX(P)=INDEX(L)
            INDEX(L)=INDEX(R)
         ELSE
            INDEX(P)=INDEX(R)
         ENDIF
         INDEX(R)=INDEXP
         INDEXP=INDEX(P)
         DATAP=DATA(INDEXP)
      ENDIF
 
!     Now we swap values between the right and left sides and/or
!     move DATAP until all smaller values are on the left and all
!     larger values are on the right.  Neither the left or right
!     side will be internally ordered yet; however, DATAP will be
!     in its final position.
 
  300 CONTINUE
 
! Q3: Search for datum on left >= DATAP
!
!     At this point, DATA[L] <= DATAP.  We can therefore start scanning
!     up from L, looking for a value >= DATAP (this scan is guaranteed
!     to terminate since we initially placed DATAP near the middle of
!     the subsequence).
 
         I=I+1
         IF (DATA(INDEX(I)).LT.DATAP) GOTO 300
 
  400 CONTINUE
 
! Q4: Search for datum on right <= DATAP
!
!     At this point, DATA[R] >= DATAP.  We can therefore start scanning
!     down from R, looking for a value <= DATAP (this scan is guaranteed
!     to terminate since we initially placed DATAP near the middle of
!     the subsequence).
 
         J=J-1
         IF (DATA(INDEX(J)).GT.DATAP) GOTO 400
 
! Q5: Have the two scans collided?
 
      IF (I.LT.J) THEN
 
! Q6: No, interchange DATA[I] <--> DATA[J] and continue
 
         INDEXT=INDEX(I)
         INDEX(I)=INDEX(J)
         INDEX(J)=INDEXT
         GOTO 300
      ELSE
 
! Q7: Yes, select next subsequence to sort
!
!     At this point, I >= J and DATA[l] <= DATA[I] == DATAP <= DATA[r],
!     for all L <= l < I and J < r <= R.  If both subsequences are
!     more than M elements long, push the longer one on the stack and
!     go back to QuickSort the shorter; if only one is more than M
!     elements long, go back and QuickSort it; otherwise, pop a
!     subsequence off the stack and QuickSort it.
 
         IF (R-J .GE. I-L .AND. I-L .GT. M) THEN
            ISTK=ISTK+1
            LSTK(ISTK)=J+1
            RSTK(ISTK)=R
            R=I-1
         ELSE IF (I-L .GT. R-J .AND. R-J .GT. M) THEN
            ISTK=ISTK+1
            LSTK(ISTK)=L
            RSTK(ISTK)=I-1
            L=J+1
         ELSE IF (R-J .GT. M) THEN
            L=J+1
         ELSE IF (I-L .GT. M) THEN
            R=I-1
         ELSE
! Q8: Pop the stack, or terminate QuickSort if empty
            IF (ISTK.LT.1) GOTO 900
            L=LSTK(ISTK)
            R=RSTK(ISTK)
            ISTK=ISTK-1
         ENDIF
         GOTO 200
      ENDIF
 
  900 CONTINUE
 
!===================================================================
!
! Q9: Straight Insertion sort
 
      DO 950 I=2,N
         IF (DATA(INDEX(I-1)) .GT. DATA(INDEX(I))) THEN
            INDEXP=INDEX(I)
            DATAP=DATA(INDEXP)
            P=I-1
  920       CONTINUE
               INDEX(P+1) = INDEX(P)
               P=P-1
               IF (P.GT.0) THEN
                  IF (DATA(INDEX(P)).GT.DATAP) GOTO 920
               ENDIF
            INDEX(P+1) = INDEXP
         ENDIF
  950    CONTINUE
 
!===================================================================
!
!     All done
 
      END	
!--------------------------------------------------------------------
!{************* MCMC runs for the Logistic regression ***************}

      SUBROUTINE MCMClogit(Y, X, n, M, burn, a, b, sigma2, tau2,&
                var_pa, var_pb, ini, Post_alpha, Post_beta, isd)
      IMPLICIT NONE
	
      INTEGER :: M, n, burn
      REAL *8 :: Y(n), X(n), ini(2)
      REAL *8 :: a, b, sigma2, tau2, var_pa, var_pb
      REAL *8 :: Post_alpha(M+burn), Post_beta(M+burn)
	
      INTEGER :: i
      REAL *8 :: ran_nor, uniform, ran_gam, ran_invgam
      REAL *8 :: alp_ini, bet_ini
      REAL *8 :: h1, h2, a_p, rho1, r1, a_new
      REAL *8 :: g1, g2, b_p, rho2, r2, b_new

      INTEGER :: is1, is2, seed1, seed2, isd
      common /unif_seeds/ is1, is2
      save /unif_seeds/
      seed1=10005
      seed2=isd  

      CALL set_uniform(seed1, seed2)	  
	  
      alp_ini = ini(1)
      bet_ini = ini(2)
      Do i = 1, (M + burn)
        a_p = ran_nor(alp_ini, var_pa)
        h1 = h(a_p, bet_ini, Y, X, n, a, sigma2)
        h2 = h(alp_ini, bet_ini, Y, X, n, a, sigma2)
        rho1 = EXP(h1 - h2)
        r1 = uniform()
        IF (r1 <= rho1) THEN 
           a_new = a_p
        ELSE 
           a_new = alp_ini
        END IF
        b_p = ran_nor(bet_ini, var_pb)
        g1 = g(a_new, b_p, Y, X, n, b, tau2)
        g2 = g(a_new, bet_ini, Y, X, n, b, tau2)
        rho2 = EXP(g1 - g2)
        r2 = uniform()
        IF (r2 <= rho2) THEN 
           b_new = b_p
        ELSE 
           b_new = bet_ini
        END IF
        Post_alpha(i) = a_new
        Post_beta(i) = b_new
        alp_ini = a_new
        bet_ini = b_new
      END DO
      RETURN
	
      CONTAINS

      Function h(alp, bet, Y, X, n, a, sigma2)
      IMPLICIT NONE

      INTEGER :: n
      REAL *8 :: alp, bet, Y(n), X(n), a, sigma2
      REAL *8 :: cmn(n)
      REAL *8 :: h

      cmn(:) = 1.0D0/(1.0D0+EXP(-alp-bet*X))
      h = SUM(Y*LOG(cmn)) + SUM((1.0D0-Y)*LOG(1.0D0-cmn)) - &
          (alp - a)**2.0D0/(2.0D0*sigma2)
      END FUNCTION h

	  
      Function g(alp, bet, Y, X, n, b, tau2)
      IMPLICIT NONE

      INTEGER :: n
      REAL *8 :: alp, bet, Y(n), X(n), b, tau2
      REAL *8 :: cmn(n)
      REAL *8 :: g

      cmn(:) = 1.0D0/(1.0D0+EXP(-alp-bet*X))
      g = SUM(Y*LOG(cmn)) + SUM((1.0D0-Y)*LOG(1.0D0-cmn)) - &
          (bet - b)**2.0D0/(2.0D0*tau2)
      END FUNCTION g

      END SUBROUTINE MCMClogit
!--------------------------------------------------------------------
!{*Posterior mean calculation for bth boostrap data using IS method*}
      SUBROUTINE ISlogit(Y, X, tab, M, n, Post_alpha, Post_beta, omega)
	  
      IMPLICIT NONE
	  
      INTEGER :: M, n
      REAL *8 :: Y(n), X(n), tab(n)
      REAL *8 :: Post_alpha(M), Post_beta(M), omega(M)
	  
      REAL *8 :: aj, bj, cmn(n)
      INTEGER :: j

	  
      DO j = 1, M
        aj = Post_alpha(j)
        bj = Post_beta(j)
        cmn(:) = 1.0D0/(1.0D0 + EXP(-aj-bj*X))
        omega(j) = EXP(SUM((tab - 1.0D0)*Y*LOG(cmn)) +&
                    SUM((tab - 1.0D0)*(1-Y)*LOG(1.0D0-cmn)))
      END DO
      RETURN

      END SUBROUTINE ISlogit
!--------------------------------------------------------------------
!{*Posterior quantile calculation for bth boostrap data using IS method*}	  
      SUBROUTINE ISlogitQtl(Y, X, tab, M, n, Post_alpha, Post_beta,&
                    probs, nprobs, qtl_alpha, qtl_beta)
      IMPLICIT NONE
	  
      INTEGER :: M, n, nprobs
      REAL *8 :: Y(n), X(n), tab(n)
      REAL *8 :: Post_alpha(M), Post_beta(M)
      REAL *8 :: aj, bj, cmn(n)
      INTEGER :: j, k, l
	  
      REAL *8 :: norm, probs(nprobs), qtl_alpha(nprobs), qtl_beta(nprobs)
      INTEGER :: alpha_idx(M), beta_idx(M), idx_tmp_alpha(nprobs), idx_tmp_beta(nprobs)
      REAL *8 :: omega(M), order_omega_alpha(M), order_omega_beta(M)
      REAL *8 :: cum_omega_alpha(M), cum_omega_beta(M)
      REAL *8 :: DF_alpha_order(M), DF_beta_order(M)

	  
      DO j = 1, M
        aj = Post_alpha(j)
        bj = Post_beta(j)
        cmn(:) = 1.0D0/(1.0D0 + EXP(-aj-bj*X))
        omega(j) = EXP(SUM((tab - 1.0D0)*Y*LOG(cmn)) +&
                    SUM((tab - 1.0D0)*(1-Y)*LOG(1.0D0-cmn)))
      END DO
      norm = SUM(omega)

	  
      DF_alpha_order = Post_alpha 
      DF_beta_order = Post_beta
      CALL SORTRX(M, DF_alpha_order, alpha_idx)
      CALL SORTRX(M, DF_beta_order, beta_idx)
      order_omega_alpha = omega(alpha_idx)
      order_omega_beta = omega(beta_idx)
	  
      DO k = 1, M
        cum_omega_alpha(k) = SUM(order_omega_alpha(1:k))/norm
		cum_omega_beta(k) = SUM(order_omega_beta(1:k))/norm
      END DO

      DO l = 1, nprobs
        idx_tmp_alpha(l) = alpha_idx(MINLOC(cum_omega_alpha, &
               DIM = 1, MASK = (cum_omega_alpha >= probs(l))))
        qtl_alpha(l) = DF_alpha_order(idx_tmp_alpha(l))
        idx_tmp_beta(l) = beta_idx(MINLOC(cum_omega_beta, &
               DIM = 1, MASK = (cum_omega_beta >= probs(l))))
        qtl_beta(l) = DF_beta_order(idx_tmp_beta(l))
      END DO
	  
      RETURN
      END SUBROUTINE ISlogitQtl
!--------------------------------------------------------------------
!{*Extreme quantile calculation for bth boostrap data using IS method*}
      SUBROUTINE ISlogitQtlextrm(Y, X, tab, a, b, sigma2, tau2, n,&
                    Post_alpha, Post_beta, M, min_a, max_a,&
                    min_b, max_b, nGL, GLw, GLx_a, eps_a,&
                    GLx_b, eps_b, min_cand_a, max_cand_a, cand_a,&
                    min_cand_b, max_cand_b, cand_b, omega_a, omega_b)
      IMPLICIT NONE

      INTEGER :: M, n, nGL
      REAL *8 :: Y(n), X(n), tab(n)
      REAL *8 :: a, b, sigma2, tau2
      REAL *8 :: Post_alpha(M), Post_beta(M)
      REAL *8 :: cand_a(M), cand_b(M)
      REAL *8 :: min_a, max_a, eps_a, h_a, min_cand_a, max_cand_a
      REAL *8 :: min_b, max_b, eps_b, h_b, min_cand_b, max_cand_b
      REAL *8 :: GLw(nGL), GLx_a(nGL), GLx_b(nGL)

      INTEGER :: j, l
      REAL *8 :: omega_a(M), cmn_a(n), cmn1_a(n), cmn2_a(n), int_tmp_a, intl_a
      REAL *8 :: omega_b(M), cmn_b(n), cmn1_b(n), cmn2_b(n), int_tmp_b, intl_b
	  
      DO j = 1, M
        cmn_a(:) = EXP(-Post_beta(j)*X)
        cmn_b(:) = EXP(-Post_alpha(j))
        cmn1_a(:) = 1/(1 + EXP(-cand_a(j))*cmn_a)
        cmn1_b(:) = 1/(1 + cmn_b*EXP(-cand_b(j)*X))
        int_tmp_a = 0
        int_tmp_b = 0
        DO l = 1, nGL
          cmn2_a(:) = 1/(1 + EXP(-GLx_a(l))*cmn_a)
          int_tmp_a = int_tmp_a + GLw(l)*&
                            PRODUCT(((cmn2_a/(cmn1_a**tab))**Y)*&
                            (((1-cmn2_a)/(1-cmn1_a)**tab)**(1-Y)))*&
               EXP((-(GLx_a(l) - a)**2 + (cand_a(j) - a)**2)/(2*sigma2))
          cmn2_b(:) = 1/(1 + cmn_b*EXP(-GLx_b(l)*X))
          int_tmp_b = int_tmp_b + GLw(l)*&
                            PRODUCT(((cmn2_b/(cmn1_b**tab))**Y)*&
                            (((1-cmn2_b)/(1-cmn1_b)**tab)**(1-Y)))*&
               EXP((-(GLx_b(l) - b)**2 + (cand_b(j) - b)**2)/(2*tau2))
        END DO
        intl_a = 1/(int_tmp_a*(max_a - min_a + 2*eps_a)/2)
        intl_b = 1/(int_tmp_b*(max_b - min_b + 2*eps_b)/2)
        h_a = 1/(max_cand_a - min_cand_a)
        h_b = 1/(max_cand_b - min_cand_b)
        omega_a(j) = intl_a/h_a
        omega_b(j) = intl_b/h_b
      END DO
	  
      RETURN
      END SUBROUTINE ISlogitQtlextrm
!--------------------------------------------------------------------
!{*****weight calculations for bth boostrap data using IS method*****}
      SUBROUTINE logitWTCoefCal(Y, X, tab, tab_b, n, Post_alpha, Post_beta,&
                    cand_a, cand_b, omega_ref_a, omega_ref_b, M, &
                    omega_new_a, omega_new_b)
      IMPLICIT NONE
	  
      INTEGER :: M, n
      REAL *8 :: Y(n), X(n), tab(n), tab_b(n)
      REAL *8 :: Post_alpha(M), Post_beta(M)
      REAL *8 :: cand_a(M), cand_b(M)
      REAL *8 :: omega_ref_a(M), omega_ref_b(M), omega_new_a(M), omega_new_b(M)

      INTEGER :: i
      REAL *8 :: cmn_a(n), cmn_b(n)
	  
      DO i = 1, M
        cmn_a = 1/(1 + EXP(-cand_a(i) - Post_beta(i)*X))
        omega_new_a(i) = omega_ref_a(i)*PRODUCT(cmn_a**((tab_b - tab)*Y)*&
                                (1-cmn_a)**((tab_b - tab)*(1-Y)))
        cmn_b = 1/(1 + EXP(-Post_alpha(i) - cand_b(i)*X))
        omega_new_b(i) = omega_ref_b(i)*PRODUCT(cmn_b**((tab_b - tab)*Y)*&
                                (1-cmn_b)**((tab_b - tab)*(1-Y)))
      END DO
      RETURN
      END SUBROUTINE logitWTCoefCal
!--------------------------------------------------------------------
!{************* MCMC runs for the Weibull regression ***************}
      SUBROUTINE MCMCWeibull(Y, D, X, a0, k0, sig0, M, burn, n, var_p, ini,& 
                isd, Post_alpha, Post_beta0, Post_beta1)
      IMPLICIT NONE

      INTEGER :: M, n, burn
      REAL *8 :: Y(n), D(n), X(n), ini(3), sumD
      REAL *8 :: a0, k0, sig0, var_p
      REAL *8 :: Post_alpha(M+burn), Post_beta0(M+burn), Post_beta1(M+burn)
	  
      INTEGER :: i
      REAL *8 :: ran_nor, uniform
      REAL *8 :: alp_ini, bet0_ini, bet1_ini, a_p, b0_p, b1_p
      REAL *8 :: rho1, rho2, rho3, r1, r2, r3
      REAL *8 :: expb0, expb1X(n)
      REAL *8 :: alp_new, bet0_new, bet1_new

      INTEGER :: is1, is2, seed1, seed2, isd
      common /unif_seeds/ is1, is2
      save /unif_seeds/
      seed1=10005
      seed2=isd  

      CALL set_uniform(seed1, seed2)
  
      sumD = SUM(D)
      alp_ini = ini(1)
      bet0_ini = ini(2)
      bet1_ini = ini(3)
      Do i = 1, (M+burn)
        expb0 = EXP(bet0_ini)
        expb1X(:) = EXP(bet1_ini*X)
        a_p = ran_nor(alp_ini, var_p)
        rho1 = EXP((a0 + sumD - 1.0D0)*(LOG(a_p) - LOG(alp_ini)) + &
               (a_p - alp_ini)*SUM(D*LOG(Y)) - &
			   SUM(expb0*expb1X*(Y**a_p - Y**alp_ini)) -&
			   k0*(a_p - alp_ini))
        r1 = uniform()
        IF (r1 <= rho1) THEN 
           alp_new = a_p
        ELSE 
           alp_new = alp_ini
        END IF
        b0_p = ran_nor(bet0_ini, var_p)
        rho2 = EXP((b0_p - bet0_ini)*sumD - (EXP(b0_p) - expb0)*&
               SUM((Y**alp_new)*expb1X) - (b0_p**2.0D0 - bet0_ini**2.0D0&
			   )/(2.0D0*10.0D0**4.0D0))
        r2 = uniform()
        IF (r2 <= rho2) THEN 
           bet0_new = b0_p
        ELSE 
           bet0_new = bet0_ini
        END IF
        b1_p = ran_nor(bet1_ini, var_p)
        rho3 = EXP((b1_p - bet1_ini)*SUM(D*X) - EXP(bet0_new)*&
               SUM((Y**alp_new)*(EXP(b1_p*X) - expb1X)) -&
               (b1_p**2.0D0 - bet1_ini**2.0D0)/(2.0D0*10.0D0**4.0D0))
        r3 = uniform()
        IF (r3 <= rho3) THEN 
           bet1_new = b1_p
        ELSE 
           bet1_new = bet1_ini
        END IF
        Post_alpha(i) = alp_new
        Post_beta0(i) = bet0_new
        Post_beta1(i) = bet1_new
        alp_ini = alp_new
        bet0_ini = bet0_new
        bet1_ini = bet1_new
      END DO
      RETURN
	  
      END SUBROUTINE MCMCWeibull
!--------------------------------------------------------------------
!{*Posterior mean calculation for bth boostrap data using IS method*}
      SUBROUTINE ISWeibull(Y, D, X, tab, a0, k0, sig0, M, n,&
                   Post_alpha, Post_beta0, Post_beta1, omega)
      IMPLICIT NONE
	  
      INTEGER :: M, n
      REAL *8 :: Y(n), D(n), X(n), tab(n)
      REAL *8 :: a0, k0, sig0
      REAL *8 :: Post_alpha(M), Post_beta0(M), Post_beta1(M)

      REAL *8 :: aj, b0j, b1j, omega(M)
      REAL *8 :: Dtab1(n), SUMDtab1, xb(n)
      INTEGER :: j
	  
      Dtab1(:) = D*(tab - 1.0D0)
      SUMDtab1 = SUM(Dtab1)
	  
      DO j = 1, M
        aj = Post_alpha(j)
        b0j = Post_beta0(j)
        b1j = Post_beta1(j)
		xb(:) = b0j + b1j*X
        omega(j) = (aj**SUMDtab1)*EXP((aj-1)*SUM(Dtab1*LOG(Y)) + &
                SUM(Dtab1*xb) - SUM(EXP(xb)*(Y**aj)*(tab-1.0D0)))
      END DO

      RETURN
      END SUBROUTINE ISWeibull
!--------------------------------------------------------------------
!{*Posterior quantile calculation for bth boostrap data using IS method*}
      SUBROUTINE ISweibullQtl(Y, D, X, tab, M, n, Post_alpha, Post_beta0, Post_beta1,&
					probs, nprobs, qtl_alpha, qtl_beta0, qtl_beta1)
      IMPLICIT NONE

      INTEGER :: M, n, nprobs
      REAL *8 :: Y(n), D(n), X(n), tab(n), probs(nprobs)
      REAL *8 :: Post_alpha(M), Post_beta0(M), Post_beta1(M)
      REAL *8 :: qtl_alpha(nprobs), qtl_beta0(nprobs), qtl_beta1(nprobs)

      INTEGER :: j, k, l
      REAL *8 :: aj, b0j, b1j, omega(M), norm
      REAL *8 :: Dtab1(n), SUMDtab1, xb(n)

      INTEGER :: alpha_idx(M), beta0_idx(M), beta1_idx(M)
      INTEGER :: idx_tmp_alpha(nprobs), idx_tmp_beta0(nprobs), idx_tmp_beta1(nprobs)
      REAL *8 :: alpha_order(M), beta0_order(M), beta1_order(M) 
      REAL *8 :: order_omega_alpha(M), order_omega_beta0(M), order_omega_beta1(M)
      REAL *8 :: cum_omega_alpha(M), cum_omega_beta0(M), cum_omega_beta1(M)


      Dtab1(:) = D*(tab - 1.0D0)
      SUMDtab1 = SUM(Dtab1)

      DO j = 1, M
        aj = Post_alpha(j)
        b0j = Post_beta0(j)
        b1j = Post_beta1(j)
		xb(:) = b0j + b1j*X
        omega(j) = aj**SUMDtab1*EXP((aj-1)*SUM(Dtab1*LOG(Y)) + &
                SUM(Dtab1*xb) - SUM(EXP(xb)*(Y**aj)*(tab-1.0D0)))
      END DO
      norm = SUM(omega)
	  
      alpha_order = Post_alpha
      beta0_order = Post_beta0
      beta1_order = Post_beta1
      CALL SORTRX(M, alpha_order, alpha_idx)
      CALL SORTRX(M, beta0_order, beta0_idx)
      CALL SORTRX(M, beta1_order, beta1_idx)
      order_omega_alpha = omega(alpha_idx)
      order_omega_beta0 = omega(beta0_idx)
      order_omega_beta1 = omega(beta1_idx)
	  
      DO k = 1, M
        cum_omega_alpha(k) = SUM(order_omega_alpha(1:k))/norm
		cum_omega_beta0(k) = SUM(order_omega_beta0(1:k))/norm
		cum_omega_beta1(k) = SUM(order_omega_beta1(1:k))/norm
      END DO

      DO l = 1, nprobs
        idx_tmp_alpha(l) = alpha_idx(MINLOC(cum_omega_alpha, &
               DIM = 1, MASK = (cum_omega_alpha >= probs(l))))
        qtl_alpha(l) = alpha_order(idx_tmp_alpha(l))
        idx_tmp_beta0(l) = beta0_idx(MINLOC(cum_omega_beta0, &
               DIM = 1, MASK = (cum_omega_beta0 >= probs(l))))
        qtl_beta0(l) = beta0_order(idx_tmp_beta0(l))
        idx_tmp_beta1(l) = beta1_idx(MINLOC(cum_omega_beta1, &
               DIM = 1, MASK = (cum_omega_beta1 >= probs(l))))
        qtl_beta1(l) = beta1_order(idx_tmp_beta1(l))
      END DO
	  
      RETURN
      END SUBROUTINE ISweibullQtl
!--------------------------------------------------------------------
!{*Extreme quantile calculation for bth boostrap data using IS method*}
      SUBROUTINE ISweibullQtlextrm(Y, D, X, tab, a0, k0, sig0, M, n,&
            Post_alpha, Post_beta0, Post_beta1, nGL, GLw, GLx_a, GLx_b0, GLx_b1,&
            min_a, max_a, eps_a, min_cand_a, max_cand_a, cand_a,&
            min_b0, max_b0, eps_b0, min_cand_b0, max_cand_b0, cand_b0,&
            min_b1, max_b1, eps_b1, min_cand_b1, max_cand_b1, cand_b1,&
            omega_a, omega_b0, omega_b1)
      IMPLICIT NONE
	  
      INTEGER :: M, n, nGL
      REAL *8 :: Y(n), D(n), X(n), tab(n)
      REAL *8 :: a0, k0, sig0, Post_alpha(M), Post_beta0(M), Post_beta1(M)
      REAL *8 :: cand_a(M), cand_b0(M), cand_b1(M)
      REAL *8 :: min_a, max_a, eps_a, min_cand_a, max_cand_a
      REAL *8 :: min_b0, max_b0, eps_b0, min_cand_b0, max_cand_b0
      REAL *8 :: min_b1, max_b1, eps_b1, min_cand_b1, max_cand_b1
      REAL *8 :: GLw(nGL), GLx_a(nGL), GLx_b0(nGL), GLx_b1(nGL)
      REAL *8 :: omega_a(M), omega_b0(M), omega_b1(M)

      INTEGER :: j, k
      REAL *8 :: eb0, eb1x(n), log_cand_a(M), log_node_a(nGL)
      REAL *8 :: D_log_Y(n), bx(n), Y_a(n)
      REAL *8 :: SumD, Sum_tab_D, cmn_a, cmn_b
      REAL *8 :: int_tmp_a, int_tmp_b0, int_tmp_b1, h_a, h_b0, h_b1
      REAL *8 :: intl_a, intl_b0, intl_b1


      D_log_Y(:) = D*LOG(Y)
      log_cand_a(:) = LOG(cand_a)
      log_node_a(:) = LOG(GLx_a)
      SumD = SUM(D)
      Sum_tab_D = SUM(tab*D)
      DO j = 1, M
        bx(:) =  Post_beta1(j)*X
        eb0 = EXP(Post_beta0(j))
        eb1x(:) = EXP(bx)
        cmn_a = SUM(D*(Post_beta0(j) + bx)*(1-tab))
        cmn_b = (SumD - Sum_tab_D)*LOG(Post_alpha(j)) +&
                SUM(D_log_Y*(1 - tab))*(Post_alpha(j) - 1)
        Y_a(:) = Y**Post_alpha(j)
        int_tmp_a = 0
        int_tmp_b0 = 0
        int_tmp_b1 = 0
        DO k = 1, nGL
            int_tmp_a = int_tmp_a + GLw(k)*&
                    EXP( SumD*log_node_a(k) - Sum_tab_D*log_cand_a(j) +&
                    (a0 - 1)*(log_node_a(k) - log_cand_a(j)) +&
                    SUM(D_log_Y*(GLx_a(k) - 1 - (cand_a(j) - 1)*tab)) -&
                    SUM(eb0*eb1x*(Y**GLx_a(k) - Y**cand_a(j))) -&
                    k0*(GLx_a(k) - cand_a(j)) + cmn_a)
            int_tmp_b0 = int_tmp_b0 + GLw(k)*&
                    EXP(SUM(D*(GLx_b0(k)-tab*cand_b0(j)+(1-tab)*bx)) -&
                    SUM(eb1x*Y_a*(EXP(GLx_b0(k))-tab*EXP(cand_b0(j))))-&
                    (GLx_b0(k)**2 - cand_b0(j)**2)/(2*(sig0**2))+cmn_b)
            int_tmp_b1 = int_tmp_b1 + GLw(k)*&
                    EXP(SUM(D*((1 - tab)*Post_beta0(j) +&
                        GLx_b1(k)*X - cand_b1(j)*tab*X)) -&
                    SUM(eb0*Y_a*(EXP(GLx_b1(k)*X) -&
                        tab*EXP(cand_b1(j)*X))) -&
                    (GLx_b1(k)**2 - cand_b1(j)**2)/(2*(sig0**2))+cmn_b)
        END DO
        intl_a = 1/(int_tmp_a*(max_a - min_a + 2*eps_a)/2)
        h_a = 1/(max_cand_a - min_cand_a)
        omega_a(j) = intl_a/h_a
        intl_b0 = 1/(int_tmp_b0*(max_b0 - min_b0 + 2*eps_b0)/2)
        h_b0 = 1/(max_cand_b0 - min_cand_b0)
        omega_b0(j) = intl_b0/h_b0
        intl_b1 = 1/(int_tmp_b1*(max_b1 - min_b1 + 2*eps_b1)/2)
        h_b1 = 1/(max_cand_b1 - min_cand_b1)
        omega_b1(j) = intl_b1/h_b1
      END DO
	  
      RETURN
      END SUBROUTINE ISweibullQtlextrm
!--------------------------------------------------------------------
!{*****weight calculations for bth boostrap data using IS method*****}	  
      SUBROUTINE WeiWTCoefCal(Y, D, X, tab, tab_b, M, n, Post_alpha,& 
            Post_beta0, Post_beta1, cand_a, cand_b0, cand_b1,&
            omega_ref_a, omega_ref_b0, omega_ref_b1, &
            omega_new_a, omega_new_b0, omega_new_b1)
      IMPLICIT NONE
	  
      INTEGER :: M, n
      REAL *8 :: Y(n), D(n), X(n), tab(n), tab_b(n)
      REAL *8 :: Post_alpha(M), Post_beta0(M), Post_beta1(M)
      REAL *8 :: cand_a(M), cand_b0(M), cand_b1(M)
      REAL *8 :: omega_ref_a(M), omega_ref_b0(M), omega_ref_b1(M)
      REAL *8 :: omega_new_a(M), omega_new_b0(M), omega_new_b1(M)
	  
      REAL *8 :: tt(n), Dtt(n), eb0, eb1x(n), bx(n), lY(n)

      INTEGER :: j
      REAL *8 :: cmn_b1, cmn_b2(n), cbx(n)

      tt(:) = tab_b - tab
      Dtt(:) = D*tt
      lY(:) = LOG(Y)
      DO j = 1, M
        bx(:) = Post_beta1(j)*X
        eb0 = EXP(Post_beta0(j))
        eb1x(:) = EXP(bx)
        omega_new_a(j) = omega_ref_a(j)*EXP(SUM(Dtt)*LOG(cand_a(j)) +&
                        SUM(Dtt*lY)*(cand_a(j) - 1) + &
                        SUM((Post_beta0(j) + bx)*Dtt) -&
                        SUM(eb0*eb1x*(Y**cand_a(j))*tt))
        cmn_b1 = SUM(Dtt)*LOG(Post_alpha(j)) +& 
                    SUM(Dtt*lY)*(Post_alpha(j) - 1)
        cmn_b2(:) = (Y**Post_alpha(j))*tt
        omega_new_b0(j) = omega_ref_b0(j)*EXP(cmn_b1 +&
                        SUM(Dtt*(cand_b0(j) + bx)) -&
                        SUM(EXP(cand_b0(j))*eb1x*cmn_b2))
        cbx(:) = cand_b1(j)*X
        omega_new_b1(j) = omega_ref_b1(j)*EXP(cmn_b1 +&
                        SUM(Dtt*(Post_beta0(j) + cbx)) -&
                        SUM(eb0*EXP(cbx)*cmn_b2))
      END DO
      RETURN 
      END SUBROUTINE WeiWTCoefCal
!--------------------------------------------------------------------
!{**** MCMC(Gibbs Sampler) for linear measurement error model. *****}
      SUBROUTINE MCMCMEM(Y, W, M, n, burn, sig2_alp, sig2_bet, sig2_mux,& 
            sig2_u, del, lam, X_ini, alp_ini, bet_ini, mux_ini,& 
            sig2e_ini, sig2x_ini, Post_alpha, Post_beta, Post_mux,& 
            Post_sig2x, Post_sig2e, Post_X, isd)
			
      IMPLICIT NONE
	
      INTEGER :: M, n, burn
      REAL *8 :: Y(n), W(n), X_ini(n), Post_X(n, M)
      REAL *8 :: ran_nor, uniform, ran_gam, ran_invgam
      REAL *8 :: sig2_alp, sig2_bet, sig2_mux, sig2_u, del, lam
      REAL *8 :: alp_ini, bet_ini, mux_ini, sig2e_ini, sig2x_ini
      REAL *8 :: Post_alpha(M), Post_beta(M), Post_mux(M)
      REAL *8 :: Post_sig2x(M), Post_sig2e(M)

      INTEGER :: i, j
      REAL *8 :: alp_new, alp_com, alp_mean, alp_var
      REAL *8 :: bet_new, bet_com, bet_mean, bet_var
      REAL *8 :: mux_new, mux_com, mux_mean, mux_var
      REAL *8 :: sig2x_new, del_x, lam_x
      REAL *8 :: sig2e_new, del_e, lam_e
      REAL *8 :: X_new(n), X_com1, X_com2, X_mean, X_var

      INTEGER :: is1, is2, seed1, seed2, isd
      common /unif_seeds/ is1, is2
      save /unif_seeds/
      seed1=10005
      seed2=isd  

      CALL set_uniform(seed1, seed2)

      DO i = 1, M+burn
        alp_com = DBLE(n) + sig2e_ini/sig2_alp
        alp_mean = SUM(Y - X_ini*bet_ini)/alp_com
        alp_var = sig2e_ini/alp_com
        alp_new = ran_nor(alp_mean, alp_var)
        bet_com = SUM(X_ini**2.0D0) + sig2e_ini/sig2_bet
        bet_mean = SUM(Y*X_ini - X_ini*alp_new)/bet_com
        bet_var = sig2e_ini/bet_com
        bet_new = ran_nor(bet_mean, bet_var)
        mux_com = DBLE(n) + sig2x_ini/sig2_mux
        mux_mean = SUM(X_ini)/mux_com
        mux_var = sig2x_ini/mux_com
        mux_new = ran_nor(mux_mean, mux_var)
        del_x = DBLE(n)/2.0D0 + del
        lam_x = SUM((X_ini - mux_new)**2)/2.0D0 + lam
        sig2x_new = ran_invgam(del_x, lam_x)
        del_e = DBLE(n)/2.0D0 + del
        lam_e = SUM((Y - alp_new - X_ini*bet_new)**2)/2.0D0 + lam
        sig2e_new = ran_invgam(del_e, lam_e)
        X_com1 = bet_new**2/sig2e_new + 1.0D0/sig2_u + 1.0D0/sig2x_new
        X_com2 = -alp_new*bet_new/sig2e_new + mux_new/sig2x_new
        X_var = 1.0D0/X_com1
        DO j = 1, n
            X_mean = (X_com2 + Y(j)*bet_new/sig2e_new +&
                      W(j)/sig2_u)/X_com1
            X_new(j) = ran_nor(X_mean, X_var)
        END DO
        Post_alpha(i) = alp_new
        Post_beta(i) = bet_new
        Post_mux(i) = mux_new
        Post_sig2x(i) = sig2x_new
        Post_sig2e(i) = sig2e_new
        Post_X(:, i) = X_new

        alp_ini = alp_new
        bet_ini = bet_new
        mux_ini = mux_new
        sig2x_ini = sig2x_new
        sig2e_ini = sig2e_new
        X_ini = X_new
      END DO
	 	  
      RETURN
      END SUBROUTINE MCMCMEM
	  
!--------------------------------------------------------------------
!{**** Importance Sampling for linear measurement error model. *****}
      SUBROUTINE ISMEM(Y, W, M, n, tab, sig2_u, Post_alpha, &
            Post_beta, Post_mux, Post_sig2x, Post_sig2e, Post_X, omega)
      IMPLICIT NONE
	  
      INTEGER :: M, n
      REAL *8 :: Y(n), W(n), Post_X(n, M), avgs(5), tab(n), sig2_u
      REAL *8 :: Post_alpha(M), Post_beta(M), Post_mux(M)
      REAL *8 :: Post_sig2x(M), Post_sig2e(M)

      INTEGER :: j
      REAL *8 :: aj, bj, muxj, sig2xj, sig2ej, Xj(n)
      REAL *8 :: omega(M), Stab1(n)
	  
      Stab1 = (tab - 1.0D0)/2.0D0

      DO j = 1, M
        aj = Post_alpha(j)
        bj = Post_beta(j)
        muxj = Post_mux(j)
        sig2xj = Post_sig2x(j)
        sig2ej = Post_sig2e(j)
        Xj(:) = Post_X(:, j)
        omega(j) = EXP(-SUM(Stab1*((Y - aj - Xj*bj)**2))/sig2ej -&
                    SUM(Stab1*((W - Xj)**2))/sig2_u -&
                    SUM(Stab1*((Xj - muxj)**2))/sig2xj)
      END DO
 
      RETURN
      END SUBROUTINE ISMEM
!--------------------------------------------------------------------
!{*Posterior quantile calculation for bth boostrap data using IS method*}
      SUBROUTINE ISMEMQtl(Y, W, M, n, tab, sig2_u, Post_alpha, Post_beta, &
                Post_mux, Post_sig2x, Post_sig2e, Post_X, probs, nprobs, &
				qtl_alpha, qtl_beta, qtl_mux, qtl_sig2x, qtl_sig2e)
      IMPLICIT NONE

      INTEGER :: M, n, nprobs
      REAL *8 :: Y(n), W(n), Post_X(n, M), tab(n), probs(nprobs)
      REAL *8 :: Post_alpha(M), Post_beta(M), Post_mux(M)
      REAL *8 :: Post_sig2x(M), Post_sig2e(M), sig2_u
      REAL *8 :: qtl_alpha(nprobs), qtl_beta(nprobs), qtl_mux(nprobs)
      REAL *8 :: qtl_sig2x(nprobs), qtl_sig2e(nprobs)	  
	  
      INTEGER :: i, j, k, l
      REAL *8 :: aj, bj, muxj, sig2xj, sig2ej, Xj(n)
      REAL *8 :: norm, omega(M), Stab1(n)
	  
      INTEGER :: alpha_idx(M), beta_idx(M), mux_idx(M)
      INTEGER :: sig2x_idx(M), sig2e_idx(M)	  
      INTEGER :: idx_tmp_alpha(nprobs), idx_tmp_beta(nprobs), idx_tmp_mux(nprobs)
      INTEGER :: idx_tmp_sig2x(nprobs), idx_tmp_sig2e(nprobs)
      REAL *8 :: alpha_order(M), beta_order(M), mux_order(M) 
      REAL *8 :: sig2x_order(M), sig2e_order(M) 
      REAL *8 :: order_omega_alpha(M), order_omega_beta(M), order_omega_mux(M)
      REAL *8 :: order_omega_sig2x(M), order_omega_sig2e(M)
      REAL *8 :: cum_omega_alpha(M), cum_omega_beta(M), cum_omega_mux(M)
      REAL *8 :: cum_omega_sig2x(M), cum_omega_sig2e(M)

      Stab1 = (tab - 1.0D0)/2.0D0

      DO j = 1, M
        aj = Post_alpha(j)
        bj = Post_beta(j)
        muxj = Post_mux(j)
        sig2xj = Post_sig2x(j)
        sig2ej = Post_sig2e(j)
        Xj(:) = Post_X(:, j)        
        omega(j) = -SUM(Stab1*((Y - aj - Xj*bj)**2))/sig2ej -&
                    SUM(Stab1*((W - Xj)**2))/sig2_u -&
                    SUM(Stab1*((Xj - muxj)**2))/sig2xj
      END DO
      omega = EXP(omega)
      norm = SUM(omega)
	  
      alpha_order = Post_alpha
      beta_order = Post_beta
      mux_order = Post_mux
      sig2x_order = Post_sig2x
      sig2e_order = Post_sig2e
      CALL SORTRX(M, alpha_order, alpha_idx)
      CALL SORTRX(M, beta_order, beta_idx)
      CALL SORTRX(M, mux_order, mux_idx)
      CALL SORTRX(M, sig2x_order, sig2x_idx)
      CALL SORTRX(M, sig2e_order, sig2e_idx)
      order_omega_alpha = omega(alpha_idx)
      order_omega_beta = omega(beta_idx)
      order_omega_mux = omega(mux_idx)
      order_omega_sig2x = omega(sig2x_idx)
      order_omega_sig2e = omega(sig2e_idx)
	  	  
      DO k = 1, M
        cum_omega_alpha(k) = SUM(order_omega_alpha(1:k))/norm
		cum_omega_beta(k) = SUM(order_omega_beta(1:k))/norm
        cum_omega_mux(k) = SUM(order_omega_mux(1:k))/norm
        cum_omega_sig2x(k) = SUM(order_omega_sig2x(1:k))/norm
        cum_omega_sig2e(k) = SUM(order_omega_sig2e(1:k))/norm
      END DO
	  
      DO l = 1, nprobs
        idx_tmp_alpha(l) = alpha_idx(MINLOC(cum_omega_alpha, &
               DIM = 1, MASK = (cum_omega_alpha >= probs(l))))
        qtl_alpha(l) = alpha_order(idx_tmp_alpha(l))
        idx_tmp_beta(l) = beta_idx(MINLOC(cum_omega_beta, &
               DIM = 1, MASK = (cum_omega_beta >= probs(l))))
        qtl_beta(l) = beta_order(idx_tmp_beta(l))
        idx_tmp_mux(l) = mux_idx(MINLOC(cum_omega_mux, &
               DIM = 1, MASK = (cum_omega_mux >= probs(l))))
        qtl_mux(l) = mux_order(idx_tmp_mux(l))
        idx_tmp_sig2x(l) = sig2x_idx(MINLOC(cum_omega_sig2x, &
               DIM = 1, MASK = (cum_omega_sig2x >= probs(l))))
        qtl_sig2x(l) = sig2x_order(idx_tmp_sig2x(l))
        idx_tmp_sig2e(l) = sig2e_idx(MINLOC(cum_omega_sig2e, &
               DIM = 1, MASK = (cum_omega_sig2e >= probs(l))))
        qtl_sig2e(l) = sig2e_order(idx_tmp_sig2e(l))
      END DO
	  
      RETURN
      END SUBROUTINE ISMEMQtl
!--------------------------------------------------------------------
!{*Extreme quantile calculation for bth boostrap data using IS method*}
      SUBROUTINE ISMEMQtlextrm(Y, W, M, n, tab, sig2_u, Post_alpha,&
            Post_beta, Post_mux, Post_sig2x, Post_sig2e, Post_X, &
            sig2_alp, sig2_bet, nGL, GLw, GLx_a, GLx_b,&
            min_a, max_a, eps_a, min_cand_a, max_cand_a, cand_a,&
            min_b, max_b, eps_b, min_cand_b, max_cand_b, cand_b,&
			omega_a, omega_b)
            
      IMPLICIT NONE

      INTEGER :: M, n, nGL
      REAL *8 :: Y(n), W(n), Post_X(n, M), tab(n)
      REAL *8 :: Post_alpha(M), Post_beta(M), Post_mux(M)
      REAL *8 :: Post_sig2x(M), Post_sig2e(M), sig2_u
      REAL *8 :: min_a, max_a, eps_a, min_cand_a, max_cand_a
      REAL *8 :: min_b, max_b, eps_b, min_cand_b, max_cand_b
      REAL *8 :: cand_a(M), cand_b(M), omega_a(M), omega_b(M)
      REAL *8 :: GLw(nGL), GLx_a(nGL), GLx_b(nGL), sig2_alp, sig2_bet


      INTEGER :: j, k
      REAL *8 :: Stab1(n), aj, candaj, bj, candbj, muxj, sig2xj, sig2ej
      REAL *8 :: Xj(n), com, com_a, com_b
      REAL *8 :: intl_a, int_tmp_a, intl_b, int_tmp_b, h_a, h_b
	  
	  
      Stab1 = (1.0D0 - tab)/2.0D0
	  
      DO j = 1, M
        aj = Post_alpha(j)
        candaj = cand_a(j)
        bj = Post_beta(j)
        candbj = cand_b(j)
        muxj = Post_mux(j)
        sig2xj = Post_sig2x(j)
        sig2ej = Post_sig2e(j)
        Xj(:) = Post_X(:, j)
        com = -SUM(Stab1*(W - Xj)**2)/sig2_u - SUM(Stab1*(Xj - muxj)**2)/sig2xj
        com_a = SUM(tab*(Y - candaj - Xj*bj)**2)/(2*sig2ej) + candaj**2/(2*sig2_alp)
        com_b = SUM(tab*(Y - aj - Xj*candbj)**2)/(2*sig2ej) + candbj**2/(2*sig2_bet)
        int_tmp_a = 0
        int_tmp_b = 0
        DO k = 1, nGL
            int_tmp_a = int_tmp_a + GLw(k)*&
                        EXP(-SUM((Y - GLx_a(k) - Xj*bj)**2)/(2*sig2ej) -&
                            GLx_a(k)**2/(2*sig2_alp) + com + com_a)
            int_tmp_b = int_tmp_b + GLw(k)*&
                        EXP(-SUM((Y - aj - Xj*GLx_b(k))**2)/(2*sig2ej) -&
                            GLx_b(k)**2/(2*sig2_bet) + com + com_b)
        END DO
        intl_a = 1/(int_tmp_a*(max_a - min_a + 2*eps_a)/2)
        h_a = 1/(max_cand_a - min_cand_a)
        omega_a(j) = intl_a/h_a
        intl_b = 1/(int_tmp_b*(max_b - min_b + 2*eps_b)/2)
        h_b = 1/(max_cand_b - min_cand_b)
        omega_b(j) = intl_b/h_b
      END DO
	  
      RETURN
      END SUBROUTINE ISMEMQtlextrm
!--------------------------------------------------------------------
!{*****weight calculations for bth boostrap data using IS method*****}	  
      SUBROUTINE WTCoefCalMEM(Y, W, M, n, tab, tab_b, sig2_u, Post_alpha,&
            Post_beta, Post_mux, Post_sig2x, Post_sig2e, Post_X, &
            cand_a, cand_b, omega_ref_a, omega_ref_b, omega_new_a, omega_new_b)
      IMPLICIT NONE
	  
	 
      INTEGER :: M, n
      REAL *8 :: Y(n), W(n), Post_X(n, M), tab(n), tab_b(n)
      REAL *8 :: Post_alpha(M), Post_beta(M), Post_mux(M)
      REAL *8 :: Post_sig2x(M), Post_sig2e(M), sig2_u
      REAL *8 :: cand_a(M), cand_b(M)
      REAL *8 :: omega_ref_a(M), omega_ref_b(M)
      REAL *8 :: omega_new_a(M), omega_new_b(M)

      INTEGER :: j
      REAL *8 :: aj, candaj, bj, candbj, muxj, sig2xj, sig2ej
      REAL *8 :: Stab1(n), com, Xj(n)
	  
      Stab1 = (tab_b - tab)/2.0D0

      DO j = 1, M
        aj = Post_alpha(j)
        candaj = cand_a(j)
        bj = Post_beta(j)
        candbj = cand_b(j)
        muxj = Post_mux(j)
        sig2xj = Post_sig2x(j)
        sig2ej = Post_sig2e(j)
        Xj(:) = Post_X(:, j)
        com = -SUM(Stab1*(W - Xj)**2)/sig2_u - SUM(Stab1*(Xj - muxj)**2)/sig2xj
        omega_new_a(j) = omega_ref_a(j)*&
                EXP(-SUM(Stab1*((Y - candaj - Xj*bj)**2))/sig2ej + com)
        omega_new_b(j) = omega_ref_b(j)*&
                EXP(-SUM(Stab1*((Y - aj - Xj*candbj)**2))/sig2ej + com)
      END DO
	  
      RETURN
      END SUBROUTINE WTCoefCalMEM